/* file : dzqueue.c
 * General purpose FIFO queue functions set.
 * Copyright (c) 1997 Dim Zegebart, Moscow Russia.
 * zager@post.comstar.ru
 * http://www.geocities.com/siliconvalley/pines/7817
 * freeware
 */

#include "palantir.h"
#include "internal.h"

//------------------------- FIFO QUEUE -----------------------------
// 'first in first out' queue functions

void queue_reset(fifo_queue *q)
{
  q->tail=0;
  q->head=0;
  q->empt=EMPTY;
}
END_OF_FUNCTION(queue_reset);

int queue_resize(fifo_queue *q,uint new_size)
{ int *tmp_mem;
  int tmp;

  lwp_disable;
  if ((tmp_mem=(int*)realloc(q->queue,sizeof(int)*new_size))==NULL) return(0);

  if (new_size>q->size) q->resize_counter++;
  else if (new_size<=q->initial_size) q->resize_counter=0;
  else q->resize_counter--;

  q->queue=tmp_mem;
  q->size=new_size;
  q->fill_level=3*(q->size>>2); // 3/4

  lwp_unlock(&q->status);
  lwp_enable;
  return(1);
}

void queue_delete(fifo_queue *q)
{ int tmp;
  lwp_disable;
  if (q==NULL)
   { lwp_enable;
     return;
   }

  free(q->queue);
  free(q);
  lwp_enable;
  return;
}

//---------------------- QUEUE EMPTY ---------------------------

inline int queue_empty(fifo_queue *q)
{ int tmp;

  lwp_disable;
  if (q==NULL)
   { lwp_enable;
     return(1);
   }
  if (q->head>=q->tail)
   { if (q->empty_handler!=NULL)
      { lwp_unlock(&q->status);
        lwp_enable;
        return(-1);
      }
     lwp_unlock(&q->status);
     lwp_enable;
     return(1);
   }
  lwp_unlock(&q->status);
  q->empt=EMPTY;
  lwp_enable;
  return(0);
}
END_OF_FUNCTION(queue_empty);

//---------------------- QUEUE PUT ----------------------------

inline int queue_put(fifo_queue *q,int c)
{ int t;
  int n=0;
  int tmp;

  lwp_disable;
  lwp_lock(&q->status);
  DISABLE();
  t=q->tail;
  q->tail++;
  q->empt=NOT_EMPTY;
  lwp_unlock(&q->status);
  ENABLE();
  if(t>=q->size)
   { q->resize_counter=0;
//     app_msg_put(system_app,SYS_QUEUE_PUT_RESET,t,q->head);
     queue_reset(q);
   }
  q->queue[t]=c;

  if (q->tail>=q->fill_level) n=1; //more then 3/4 of queue
  lwp_enable;
  return(n);
}
END_OF_FUNCTION(queue_put);

//---------------------- QUEUE GET ---------------------------

inline int queue_get(fifo_queue *q)
{ int h;
  int c;
  int tmp;

  lwp_disable;
  lwp_lock(&q->status);
  DISABLE();
  h=q->head;
  q->head++;
  lwp_unlock(&q->status);
  ENABLE();
  c=q->queue[h];
  if (q->head>=q->tail) queue_reset(q);
  lwp_enable;
  return(c);
}
END_OF_FUNCTION(queue_get);

/*
inline void queue_get_n(fifo_queue *q,int *mem_to,byte n,get_mode m)
{ DISABLE();
  if (n!=0) memcpy((void*)mem_to,(void*)&(q->queue[q->head]),sizeof(int)*n);
  else memcpy((void*)mem_to,(void*)(&(q->queue[q->head])),sizeof(int)*(q->tail-q->head));
//  if (m==NORMAL)
   { q->head+=n;
     if (queue_empty(q)) queue_reset(q);
   }
  ENABLE();
}

inline int queue_put_n(fifo_queue *q,int *mem_from,byte n)
{ int k=0;
  DISABLE();
  if(q->tail>=(q->size-n-1))
   { q->resize_counter=0;
     #ifdef __DEVICE__
     system_msg_put(SYS_QUEUE_PUT_N_RESET,q->size,q->head);
     #endif
     queue_reset(q);
   }

  memcpy((void*)(&(q->queue[q->tail])),(void*)mem_from,sizeof(int)*n);
  q->tail+=n;
  if (q->tail>=q->fill_level) k=1; //more the 7/16 of queue
  ENABLE();
  return(k);
}
*/

fifo_queue* queue_new(uint size)
{ fifo_queue *q;
  int tmp;

  lwp_disable;
  q=(fifo_queue*)malloc(sizeof(fifo_queue));
  lwp_enable;
  if (q==NULL) return(NULL);

  lwp_disable;
  if (!size) size=1024; //if illegal size, turn it to default size
  if ((q->queue=(int*)malloc(sizeof(int)*size))==NULL) return(NULL);

  lwp_lock(&q->status);
  q->size=size;
  q->initial_size=size;
  q->resize_counter=0;
  q->fill_level=(3*q->size)/4; // 3/4
  q->empty_handler=NULL;
  queue_reset(q);

  _go32_dpmi_lock_data(q->queue,sizeof(int)*size);
  _go32_dpmi_lock_data(q,sizeof(fifo_queue));

  lwp_enable;
  lwp_unlock(&q->status);
  return(q);
}

void queue_init()
{
  LOCK_FUNCTION(queue_empty);
  LOCK_FUNCTION(queue_put);
  LOCK_FUNCTION(queue_get);
  LOCK_FUNCTION(queue_reset);
}